#include "json.hpp"
#include "enums.hpp"
#include <string>
#include <vector>
#include <unordered_map>
#include <iostream>
using namespace std;
using json = nlohmann::json;

// ak napr. x = 3, dmp(x) vypise "x = 3" na stderr,
// mozete zmazat ak sa nepaci :))
#define dmp(x) cerr << #x << " = " << x << endl


/* Táto trieda reprezentuje pozíciu políčka v šesťuholníkovej sieti.
Vieme ich sčitovať a počítať medzi nimi vzdialenosť. */
struct Position {
    int q, r, s;

    Position() = default;

    Position(int q, int r, int s) {
        this->q = q;
        this->r = r;
        this->s = s;
    }

    Position(json& data) {
        // Position(data.at("q"), data.at("r"), data.at("s"));
        this->q = data.at("q");
        this->r = data.at("r");
        this->s = data.at("s");
    }

    Position operator+(const Position& other) const {
        return Position(q + other.q,
                        r + other.r,
                        s + other.s);
    }

    Position operator-(const Position& other) const {
        return Position(q - other.q,
                        r - other.r,
                        s - other.s);
    }

    bool operator==(const Position& other) const {
        return (q == other.q &&
                r == other.r &&
                s == other.s);
    }

    int distance(const Position& other) {
        return (abs(q - other.q) +
                abs(r - other.r) +
                abs(s - other.s)) / 2;
    }

    json to_json() {
        return {
            {"q", q},
            {"r", r},
            {"s", s}
        };
    }

    friend std::ostream& operator<<(std::ostream& os, const Position& pos) {
        os << "Position(" << pos.q << ", " << pos.r << ", " << pos.s << ")";
        return os;
    }
};

namespace std {
    template<>
    struct hash<Position> {
        const int BASE = 1e6;
        const int MOD = 1e9 + 7;

        int add(int x, int y) const {
            return (x + y) % MOD;
        }

        int mul(int x, int y) const {
            return ((long long)x * (long long)y) % (long long)MOD;
        }

        size_t operator()(const Position& pos) const {
            int h1 = hash<int>{}(pos.q) + 1;
            int h2 = hash<int>{}(pos.r) + 1;
            int h3 = hash<int>{}(pos.s) + 1;
            int x1 = mul(h1, mul(BASE, BASE));
            int x2 = mul(h2, BASE);
            return add(x1, add(x2, h3));
        }
    };
}

/* Vracia opacny smer. */
inline Directions opposite_dir(Directions direction) {
    if (direction == Directions::DirectionNONE) return direction;
    int idx = static_cast<int>(direction);
    return static_cast<Directions>((idx + 3) % 6);
}

inline vector<Position> dirs_to_pos = {
    Position(1, -1, 0),
    Position(1, 0, -1),
    Position(0, 1, -1),
    Position(-1, 1, 0),
    Position(-1, 0, 1),
    Position(0, -1, 1)
};

inline Directions pos_to_dir(Position pos) {
    if (pos == Position(1, -1, 0)) {
        return Directions::DirectionNE;
    } else if (pos == Position(1, 0, -1)) {
        return Directions::DirectionE;
    } else if (pos == Position(0, 1, -1)) {
        return Directions::DirectionSE;
    } else if (pos == Position(-1, 1, 0)) {
        return Directions::DirectionSW;
    } else if (pos == Position(-1, 0, 1)) {
        return Directions::DirectionW;
    } else if (pos == Position(0, -1, 1)) {
        return Directions::DirectionNW;
    }
    return Directions::DirectionNONE;
}

/* Trieda Player reprezentuje nášho hráča. Nájdeme v nej naše ID a aktuálne
 peniaze(updatujú sa na začiatku každého kola). */
struct Player {
    int ID;
    int money;
    Player() = default; 

    Player(int ID, int money) {
        this->ID = ID;
        this->money = money;
    }

    friend std::ostream& operator<<(std::ostream& os, const Player& player) {
        os << "Player(" << player.ID << ", " << player.money << ")";
        return os;
    }
};

/* Tile je generická trieda pre políčko.
Políčko má pozíciu a vlastníka práva na výhybku. */
struct Tile {
    Position position;
    // Ak je -1, tak nema nikto.
    int right_of_switch = -1;

    Tile() = default;

    Tile(Position position) {
        this->position = position;
    }

    /* Vracia true ak sa dá na políčku stavať. */
    virtual bool buildable(Player player) {
        throw logic_error("Not implemented");
    }

    /* Vracia maximálny počet vlakov, ktoré môžu byť na políčku. */
    virtual int max_trains() {
        return 0;
    }

    /* Vracia susedné políčko vo vybranom smere. */
    Tile* map_neighbour(unordered_map<Position, Tile*> &grid, Directions direction) {
        int idx = static_cast<int>(direction);
        Position pos = dirs_to_pos.at(static_cast<int>(direction));
        if (grid.find(position + pos) == grid.end()) {
            return new Tile(Position(INT_MAX, INT_MAX, INT_MAX));
        }
        return grid.at(position + pos);
    }

    /* Vracia vektor všetkých susedných políčok. */
    vector<Tile*> all_neighbours(unordered_map<Position, Tile*> &grid) {
        vector<Tile*> neighbours;
        for (Position pos : dirs_to_pos) {
            if (grid.find(position + pos) != grid.end()) {
                neighbours.push_back(grid.at(position + pos));
            }
        }
        return neighbours;
    }
    /* Vracia vektor všetkých susedných políčok, na ktoré sa váš vlak dokáže dostať,
     * keď prichádza z vybraného smeru. */
    virtual vector<Tile*> neighbours_from_direction(Directions direction,
            unordered_map<Position, Tile*>& grid) {
        return {};
    }

    /* Vracia 3 policka na mape ktore su oproti danemu smeru. Vracia len existujuce policka z mapy. */
    vector<Tile*> opposite_triplet_from_dir(unordered_map<Position, Tile*> &grid,
        Directions direction) {
        vector<Directions> dirs = { opposite_dir(direction) };
        dirs.push_back(static_cast<Directions>((static_cast<int>(dirs[0]) + 1) % 6));
        dirs.push_back(static_cast<Directions>((static_cast<int>(dirs[0]) + 5) % 6));

        vector<Tile*> res;
        for (Directions dir : dirs) {
            Position ngb_pos = dirs_to_pos.at(static_cast<int>(dir));
            if (grid.find(ngb_pos) != grid.end()) {
                res.push_back(grid.at(ngb_pos));
            }
        }
        return res;
    }

    /* Vracia smer v ktorom je vybrané susedné políčko. */
    Directions dir_to_neighbour(Tile* other) {
        Position diff = other->position - position;
        return pos_to_dir(diff);
    }

    /* Vracia množinu susedných políčok, na ktoré sa dá z políčka dostať po trati, bez ohľadu na smer. */
    virtual vector<Tile*> reachable(unordered_map<Position, Tile*> &grid) {
        return all_neighbours(grid);
    }

    protected: static vector<Tile*> tiles_to_vec(vector<Tile*> tiles) {
        vector<Tile*> res;
        for (Tile* tile : tiles) {
            if (tile->position.q != INT_MAX) {
                res.push_back(tile);
            } else {
                delete tile;
            }
        }
        return res;
    }
};

struct TileEmpty : Tile {
    TileEmpty(Position position) :
        Tile(position) {}
    
    bool buildable(Player player) override {
        return true;
    }

    int max_trains() override {
        return 0;
    }

    vector<Tile*> reachable(unordered_map<Position, Tile*> &grid) override {
        return {};
    }

    friend std::ostream& operator<<(std::ostream& os, const TileEmpty& tile) {
        os << "TileEmpty(" << tile.position << ")";
        return os;
    }
};

struct TileWater : Tile {
    TileWater(Position position) :
        Tile(position) {}
    
    bool buildable(Player player) override {
        return false;
    }

    int max_trains() override {
        return 1;
    }

    friend std::ostream& operator<<(std::ostream& os, const TileWater& tile) {
        os << "TileWater(" << tile.position << ")";
        return os;
    }
};

struct TileTrack : Tile {
    pair<Directions, Directions> segments;
    TileTrack(Position position, int player_id, pair<Directions, Directions> segments) : Tile(position) {
        this->right_of_switch = player_id;
        this->segments = segments;
    }
    
    bool buildable(Player player) override {
        return right_of_switch == player.ID;
    }

    int max_trains() override {
        return 1;
    }

    vector<Tile*> neighbours_from_direction(Directions direction,
            unordered_map<Position, Tile*> &grid) override {
        if (opposite_dir(direction) == segments.first) {
            return tiles_to_vec({ map_neighbour(grid, segments.second) });
        } else if (opposite_dir(direction) == segments.second) {
            return tiles_to_vec({ map_neighbour(grid, segments.first) });
        }
        return {};
    }

    vector<Tile*> reachable(unordered_map<Position, Tile*> &grid) override {
        return tiles_to_vec(
                { map_neighbour(grid, segments.first),  map_neighbour(grid, segments.second) }
        );
    }

    friend std::ostream& operator<<(std::ostream& os, const TileTrack& tile) {
        os << "TileTrack(" << tile.position << ", " << tile.right_of_switch << ", " << tile.segments << ")";
        return os;
    }
};

struct TileSwitch : Tile {
    bool is_switched;
    Directions straight_direction;
    Directions switched_direction;
    TileSwitch(Position position, int player_id, bool is_switched, Directions straight_direction,
            Directions switched_direction) : Tile(position) {
        this->right_of_switch = player_id;
        this->is_switched = is_switched;
        this->straight_direction = straight_direction;
        this->switched_direction = switched_direction;

    }

    bool buildable(Player player) override {
        return right_of_switch == player.ID;
    }

    int max_trains() override {
        return 1;
    }

    vector<Tile*> neighbours_from_direction(Directions direction,
            unordered_map<Position, Tile*> &grid) override {
        if (direction == straight_direction) {
            return tiles_to_vec({
                map_neighbour(grid, switched_direction),
                map_neighbour(grid, opposite_dir(switched_direction))
            });
        } else if (direction == opposite_dir(straight_direction)) {
            return tiles_to_vec({ map_neighbour(grid, direction) });
        } else if (direction == switched_direction && is_switched) {
            return tiles_to_vec({ map_neighbour(grid, opposite_dir(direction)) });
        }
        return {};
    }

    vector<Tile*> reachable(unordered_map<Position, Tile*> &grid) override {
        return tiles_to_vec({
            map_neighbour(grid, straight_direction),
            map_neighbour(grid, opposite_dir(straight_direction)),
            map_neighbour(grid, switched_direction)
        });
    }

    friend std::ostream& operator<<(std::ostream& os, const TileSwitch& tile) {
        os << "TileTrack(" << tile.position << ", " << tile.is_switched << ", " <<
            tile.straight_direction << ", " << tile.switched_direction << ")";
        return os;
    }
};

struct TileCrossing : Tile {
    pair<Directions, Directions> directions;
    TileCrossing(Position position, int player_id,
            pair<Directions, Directions> directions) : Tile(position) {
        this->right_of_switch = player_id;
        this->directions = directions;
    }

    bool buildable(Player other) override {
        return right_of_switch == other.ID;
    }

    int max_trains() override {
        return 1;
    }

    vector<Tile*> neighbours_from_direction(Directions direction,
            unordered_map<Position, Tile*> &grid) override {
        if (direction == directions.first) {
            return tiles_to_vec({ map_neighbour(grid, direction) });
        } else if (direction == directions.second) {
            return tiles_to_vec({ map_neighbour(grid, direction) });
        } else if (direction == opposite_dir(directions.first)) {
            return tiles_to_vec({ map_neighbour(grid, direction) });
        } else if (direction == opposite_dir(directions.second)) {
            return tiles_to_vec({ map_neighbour(grid, direction) });
        }
        return {};
    }

    vector<Tile*> reachable(unordered_map<Position, Tile*> &grid) override {
        return tiles_to_vec({
            map_neighbour(grid, directions.first),
            map_neighbour(grid, opposite_dir(directions.first)),
            map_neighbour(grid, directions.second),
            map_neighbour(grid, opposite_dir(directions.second))
        });
    }

    friend std::ostream& operator<<(std::ostream& os, const TileCrossing& tile) {
        os << "TileCrossing(" << tile.position << ", " << tile.directions << ")";
        return os;
    }
};

struct TileDepot : Tile {
    public: TileDepot() = default;

    TileDepot(Position position, int player_id) : Tile(position) {
        this->right_of_switch = player_id;
    }

    bool buildable(Player player) override {
        return false;
    }

    int max_trains() override {
        return 1e9 + 7;
    }

    vector<Tile*> neighbours_from_direction(Directions direction,
            unordered_map<Position, Tile*> &grid) override {
        return { tiles_to_vec({ map_neighbour(grid, direction) }) };
    }

    friend std::ostream& operator<<(std::ostream& os, const TileDepot& tile) {
        os << "TileDepot(" << tile.position << ", " << tile.right_of_switch << ")";
        return os;
    }
};


struct TileFactory : Tile {
    Directions straight_direction;
    FactoryType type;
    unordered_map<Materials, int> cargo;

    public: TileFactory() = default;

    TileFactory(Position position, Directions straight_direction, FactoryType type,
            unordered_map<Materials, int> cargo) : Tile(position) {
        this->straight_direction = straight_direction;
        this->type = type;
        this->cargo = cargo;
    }

    bool buildable(Player player) override {
        return false;
    }

    int max_trains() override {
        return 1e9 + 7;
    }

    vector<Tile*> neighbours_from_direction(Directions direction,
            unordered_map<Position, Tile*> &grid) override {
        if (direction == straight_direction) {
            return tiles_to_vec({ map_neighbour(grid, straight_direction) });
        } else if (direction == opposite_dir(straight_direction)) {
            return tiles_to_vec({ map_neighbour(grid, opposite_dir(straight_direction)) });
        }
        return {};
    }

    vector<Tile*> reachable(unordered_map<Position, Tile*> &grid) override {
        return tiles_to_vec({
            map_neighbour(grid, straight_direction),
            map_neighbour(grid, opposite_dir(straight_direction))
        });
    }

    friend std::ostream& operator<<(std::ostream& os, const TileFactory& tile) {
        os << "TileFactory(" << tile.position << ", " << tile.straight_direction << ", " << tile.type << ")";
        return os;
    }
};

/* Train je trieda, ktorá reprezentuje Vlak.

O vlakoch vieme povedať: ich id, vlastníka, pozíciu, smer, počet vagónov,
prevážanú surovinu a jej množstvo. */
struct Train {
    int id;
    int player;
    Position position;
    Directions direction;
    int wagons;
    int cargo_amount;
    Materials cargo_material;

    Train(json& data) {
        id = data.at("id");
        player = data.at("player");
        position = Position(data.at("position"));
        direction = static_cast<Directions>(data.at("direction"));
        wagons = data.at("wagons");
        cargo_amount = data.at("cargo_amount");
        cargo_material = static_cast<Materials>(data.at("cargo_material"));
    }

    friend std::ostream& operator<<(std::ostream& os, const Train& train) {
        os << "Train(" << train.id << ", " << train.position << ", " << train.direction <<
            ", " << train.wagons << ", " << train.cargo_amount << ")";
        return os;
    }
};

struct Turn {
    virtual string get_name() {
        return "";
    }

    virtual json get_data() {
        return {};
    }

    virtual json to_json() {
        return {
            {"turn", get_name()},
            {"data", get_data()},
        };
    }
};

/* TurnBuild reprezentuje ťah stavania trate. */
struct TurnBuild : Turn {
    string get_name() override {
        return "build";
    }

    Position position;
    vector<int> segments;
    /* Pre stavanie potrebujeme zvoliť pozíciu políčka na mape a zoznam segmentov, ktorý chceme stavať. */
    TurnBuild(Position position, vector<Directions> segments) : position(position){
        for(Directions dir : segments){
            this->segments.push_back(int(dir));
        }
    };

    json get_data() override {
        return {
            {"position", position.to_json()},
            {"segment", static_cast<vector<int>>(segments)}
        };
    }

    friend std::ostream& operator<<(std::ostream& os, const TurnBuild& turn) {
        os << "TurnBuild(" << turn.position << ", " << turn.segments << ")";
        return os;
    }
};

/* TurnBuild reprezentuje ťah stavania depa. */
struct TurnBuildDepot : Turn {
    string get_name() override {
        return "build_depot";
    }

    Position position;
    /* Pre stavanie potrebujeme zvoliť pozíciu políčka na mape. */
    TurnBuildDepot(Position &position) : position(position) {};

    json get_data() override {
        return { {"position", position.to_json()} };
    }

    friend std::ostream& operator<<(std::ostream& os, const TurnBuildDepot& turn) {
        os << "TurnBuildDepot(" << turn.position << ")";
        return os;
    }
};

/* TurnToggleSwitch reprezentuje ťah prehodenie výhybku. */
struct TurnToggleSwitch : Turn {
    string get_name() override {
        return "toggle_switch";
    }

    Position position;
    /* Na prehodenie výhybky potrebe zvoliť pozíciu políčka s výhybkou. */
    TurnToggleSwitch(Position position) :
        position(position) {};

    friend std::ostream& operator<<(std::ostream& os, const TurnToggleSwitch& turn) {
        os << "TurnToggleSwitch(" << turn.position << ")";
        return os;
    }
};

/* TurnDestroy reprezentuje ťah zničenia trate. Zničí všetko a na políčku bude EmptyTile.
 * Všetky vlaky zničí tiež. */
struct TurnDestroy : Turn {
    string get_name() override {
        return "destroy";
    }

    Position position;
    /* Na zničenie potrebujeme zvoliť pozíciu políčka. */
    TurnDestroy(Position position) :
        position(position) {};

    json get_data() override {
        return { {"position", position.to_json()} };
    }

    friend std::ostream& operator<<(std::ostream& os, const TurnDestroy& turn) {
        os << "TurnDestroy(" << turn.position << ")";
        return os;
    }
};

/* TurnTrainDirection reprezentuje ťah zmeny smeru vlaku. Taktiež ide začatie pohybu.
 * Za jazdy smer však meniť nevieme. */
struct TurnTrainDirection : Turn {
    string get_name() override {
        return "train_direction";
    }

    Train train;
    Directions direction;
    /* Pre zmenu smeru vlaku potrebujeme zvoliť vlak a smer, do ktorého chceme aby šiel. */
    TurnTrainDirection(Train train, Directions direction) :
        train(train), direction(direction) {};

    json get_data() override {
        return {
            {"train", train.id},
            {"direction", static_cast<int>(direction)}
        };
    }

    friend std::ostream& operator<<(std::ostream& os, const TurnTrainDirection& turn) {
        os << "TurnTrainDirection(" << turn.train << ", " << turn.direction << ")";
        return os;
    }
};

/* TurnToggleSwitch reprezentuje ťah zastavenia vlaku. */
struct TurnTrainStop : Turn {
    string get_name() override {
        return "train_stop";
    }

    Train train;
    /* Na zastavenie vlaku potrebujeme zvoliť vlak, ktorý chceme zastaviť. */
    TurnTrainStop(Train train) :
        train(train) {};

    json get_data() override {
        return {
            {"train", train.id}
        };
    }

    friend std::ostream& operator<<(std::ostream& os, const TurnTrainStop& turn) {
        os << "TurnTrainStop(" << turn.train << ")";
        return os;
    }
};

/* TurnTrainLoad reprezentuje ťah nakládky vlaku. Náš vlak pri tom musí stáť vo fabrike. */
struct TurnTrainLoad : Turn {
    string get_name() override {
        return "train_load";
    }

    Train train;
    int cargo_amount;
    Materials cargo_material;
    /* Pre vykládku potrebujeme zvoliť vlak, ktorý sa má naložiť, počet suroviny,
     * ktorú chceme vo fabrike vyložiť a taktiež o akú surovinu ide. */
    TurnTrainLoad(Train train, int cargo_amount, Materials cargo_material) :
        train(train), cargo_amount(cargo_amount), cargo_material(cargo_material) {};

    json get_data() override {
        return {
            {"train", train.id},
            {"cargo_amount", cargo_amount},
            {"cargo_material", static_cast<int>(cargo_material)}
        };
    }

    friend std::ostream& operator<<(std::ostream& os, const TurnTrainLoad& turn) {
        os << "TurnTrainLoad(" << turn.train << ", " << turn.cargo_amount << ", " << turn.cargo_material << ")";
        return os;
    }
};

/* TurnTrainUnload reprezentuje ťah vykládky vlaku. Náš vlak pri tom musí stáť vo fabrike. */
struct TurnTrainUnload : Turn {
    string get_name() override {
        return "train_unload";
    }

    Train train;
    int cargo_amount;
    /* Pre vykládku potrebujeme zvoliť vlak, ktorý sa má vyložiť a počet suroviny,
     * ktorú chceme vo fabrike vyložiť.*/
    TurnTrainUnload(Train train, int cargo_amount) :
        train(train), cargo_amount(cargo_amount) {};

    json get_data() override {
        return {
            {"train", train.id},
            {"cargo_amount", cargo_amount}
        };
    }

    friend std::ostream& operator<<(std::ostream& os, const TurnTrainUnload& turn) {
        os << "TurnTrainUnload(" << turn.train << ", " << turn.cargo_amount << ")";
        return os;
    }
};

/* TurnTrainBuy reprezentuje ťah nákupu vlaku. */
struct TurnTrainBuy : Turn {
    string get_name() override {
        return "train_buy";
    }

    Position depot;
    /* Pre nákup vlaku potrebujeme zvoliť niektoré naše depo, v ktorom sa vytvorí. */
    TurnTrainBuy(Position depot) :
        depot(depot) {};

    json get_data() override {
        return {
            {"depot", depot.to_json()}
        };
    }

    friend std::ostream& operator<<(std::ostream& os, const TurnTrainBuy& turn) {
        os << "TurnTrainBuy(" << turn.depot << ")";
        return os;
    }
};

/* TurnTrainAddWagon reprezentuje ťah nákupu vagónu. Vlak pri tom musí stáť niektorom našom depe. */
struct TurnTrainAddWagon : Turn {
    string get_name() override {
        return "train_add_wagon";
    }

    Train train;
    /* Pre nákup vagónu musíme zvoliť vlak, do ktorého chceme pridať vagón. */
    TurnTrainAddWagon(Train train) :
        train(train) {};

    json get_data() override {
        return {
            {"train", train.id}
        };
    }

    friend std::ostream& operator<<(std::ostream& os, const TurnTrainAddWagon& turn) {
        os << "TurnTrainAddWagon(" << turn.train << ")";
        return os;
    }
};

/* TurnHorn reprezentuje ťah trúbenia vlaku. Máme na výber z dvoch vlastných mp3. */
struct TurnHorn : Turn {
    string get_name() override {
        return "horn";
    }

    int id;
    /* Na zatrúbenie potrebujeme zvoliť id mp3ky. */
    TurnHorn(int id) :
        id(id) {};

    json get_data() override {
        return {
            {"id", id}
        };
    }

    friend std::ostream& operator<<(std::ostream& os, const TurnHorn& turn) {
        os << "TurnHorn(" << turn.id << ")";
        return os;
    }
};

/* Game je trieda, ktorá preprezenuje celú hru.

* map - objekt Mapy

* player - váš hráč

* enemy_player - zoznam neprateľských hráčov, ktorých vidíte

* items - zoznam všetkých vecí, ktoré vidíte */
struct Game {
    Player player;
    string color;
    unordered_map<Position, Tile*> grid;
    int radius;
    vector<Train> trains;
    vector<Train> my_trains;
    unordered_map<Position, TileDepot*> my_depos;
    unordered_map<Position, TileFactory*> factories;

    Game(string color) {
        this->color = color;
        grid = {};
        trains = {};
        my_trains = {};
        trains = {};
        my_depos = {};
        factories = {};
    }

    void fill_with_empty() {
        if (grid.size() > radius) {
            return;
        }
        for (int r = -radius; r <= radius; r++) {
            int min_q = (r < 0) ? -radius - r : -radius;
            int max_q = (r > 0) ? radius - r : radius;
            for (int q = min_q; q <= max_q; q++) {
                int s = -q - r;
                Position pos(q, r, s);
                grid[pos] = new TileEmpty(pos);
            }
        }
    }

    private : bool read_state(int round) {
        ostringstream oss;
        string line;
        while (getline(cin, line)) {
            if (line == ".") {
                break;
            }
            if (line.empty()) {
                continue;
            }
            oss << line << '\n';
        }

        if (oss.str().empty()) {
            return false;
        }
        json data = json::parse(oss.str());
        my_trains.clear();
        trains.clear();

        radius = data.at("map_radius");
        player = Player(data.at("player").at("id"), data.at("player").at("money"));
        if (round == 0) {
            fill_with_empty();
        }

        for (json& tile_data : data.at("dirty_tiles")) {
            string type = tile_data.at("type");
            json& tile = tile_data.at("data");
            Position pos(tile.at("position"));

            int player_id = -1;
            if (!tile.at("right_of_switch").is_null()) {
                player_id = tile.at("right_of_switch");
            }
            if (grid.find(pos) != grid.end()) {
                Tile *t = grid.at(pos);
                if (dynamic_cast<TileFactory*>(t)) {
                    factories.erase(t->position);
                } else if (TileDepot *td = dynamic_cast<TileDepot*>(t)) {
                    if (td->right_of_switch == player_id) {
                        my_depos.erase(t->position);
                    }
                }
                delete t;
            }
            if (type == "track") {
                grid[pos] = new TileTrack(pos, player_id, tile.at("segments"));
            } else if (type == "switch") {
                grid[pos] = new TileSwitch(pos, player_id, tile.at("is_switched"),
                    tile.at("straight_direction"), tile.at("switched_direction"));
            } else if (type == "crossing") {
                grid[pos] = new TileCrossing(pos, player_id, tile.at("directions"));
            } else if (type == "depot") {
                TileDepot* depot = new TileDepot(pos, player_id);
                grid[pos] = depot;
                if (player.ID == player_id) {
                    my_depos[pos] = depot;
                }
            } else if (type == "factory") {
                json& cargo = tile.at("cargo");
                unordered_map<Materials, int> m;
                for (auto& [key, value] : cargo.items()) {
                    m[static_cast<Materials>(key[0] - '0')] = value;
                }
                FactoryType ft = static_cast<FactoryType>(tile.at("factory_type"));
                Directions dir = static_cast<Directions>(tile.at("straight_direction"));
                TileFactory* fact = new TileFactory(pos, dir, ft, m);
                grid[pos] = fact;
                factories[pos] = fact;
            } else if (type == "empty") {
                grid[pos] = new TileEmpty(pos);
            } else if (type == "water") {
                grid[pos] = new TileWater(pos);
            }
        }
        for (json& train : data.at("trains")) {
            Train t(train);
            if (t.player == player.ID) {
                my_trains.push_back(t);
            }
            trains.push_back(t);
        }
        return true;
    }

    void greet() {
        json j = { {"color", color} };
        cout << j.dump() << endl;
        cout << "." << endl;
    }

    virtual vector<Turn*> make_turns() {
        throw logic_error("Not implemented");
    }

    void send_turns(vector<Turn*>& turns) {
        vector<json> _turns;
        for (Turn* turn : turns) {
            _turns.push_back(turn->to_json());
        }
        json j = _turns;
        cout << j.dump() << endl;
        cout << "." << endl;
        for (Turn* turn : turns) {
            delete turn;
        }
    }

    public : void run() {
        greet();
        for (int round = 0; ; round++) {
            if (!read_state(round)) {
                break;
            }
            vector<Turn*> turns = make_turns();
            send_turns(turns);
        }
        for (auto pr : grid) {
            delete pr.second;
        }
    }
};
