import enum
import json
import sys
from typing import Tuple, Union


class Materials(enum.IntEnum):
    """
    Táto trieda obsahuje typy surovín, ktoré môžete prevážať.
    """

    IronOre = 0
    Coal = 1
    Logs = 2
    Grain = 3
    Steel = 4
    Planks = 5
    Food = 6
    Goods = 7


class Position:
    """
    Táto trieda reprezentuje pozíciu políčka v šesťuholníkovej sieti.
    Vieme ich sčitovať a počítať medzi nimi vzdialenosť.
    """

    def __init__(self, q: int, r: int, s: int):
        self.q = q
        self.r = r
        self.s = s

    @classmethod
    def from_json(cls, data: dict):
        return cls(data["q"], data["r"], data["s"])

    def __add__(self, other: "Position") -> "Position":
        return Position(self.q + other.q, self.r + other.r, self.s + other.s)

    def __sub__(self, other: "Position") -> "Position":
        return Position(self.q - other.q, self.r - other.r, self.s - other.s)

    def distance(self, other: "Position") -> float:
        """
        Metóda na výpočet Manhattan vzdialenosť medzi dvoma bodmi.
        params:
            - other: Position - k akéj pozícii počítane vzdialenosť
        """
        return (
                abs(self.q - other.q) + abs(self.r - other.r) + abs(self.s - other.s)
        ) / 2

    def to_dict(self):
        return {"q": self.q, "r": self.r, "s": self.s}

    def __hash__(self):
        return hash((self.q, self.r, self.s))

    def __eq__(self, other):
        return self.q == other.q and self.r == other.r and self.s == other.s

    def __str__(self):
        return f"Position({self.q}, {self.r}, {self.s})"

    def __repr__(self):
        return self.__str__()


class Directions(enum.IntEnum):
    """Táto trieda reprezentuje smery"""

    NE = 0
    E = 1
    SE = 2
    SW = 3
    W = 4
    NW = 5
    NONE = -1

    def opposite(self):
        """Metóda vracia opačný smer"""
        if self.value == -1:
            return Directions(-1)
        return Directions((self.value + 3) % 6)


_direction_positions = {
    Directions.NE: Position(q=1, r=-1, s=0),
    Directions.E: Position(q=1, r=0, s=-1),
    Directions.SE: Position(q=0, r=1, s=-1),
    Directions.SW: Position(q=-1, r=1, s=0),
    Directions.W: Position(q=-1, r=0, s=1),
    Directions.NW: Position(q=0, r=-1, s=1),
}

_position_directions = {
    Position(q=1, r=-1, s=0): Directions.NE,
    Position(q=1, r=0, s=-1): Directions.E,
    Position(q=0, r=1, s=-1): Directions.SE,
    Position(q=-1, r=1, s=0): Directions.SW,
    Position(q=-1, r=0, s=1): Directions.W,
    Position(q=0, r=-1, s=1): Directions.NW,
}


def _tiles_to_set(tiles: list[Union["Tile", None]]) -> set["Tile"]:
    return {tile for tile in tiles if tile is not None}


class Player:
    """
    Trieda Player reprezentuje nášho hráča. Nájdeme v nej naše ID a aktuálne peniaze(updatujú sa na začiatku každého kola).
    """

    def __init__(self, ID: int, money: int):
        self.ID = ID
        self.money = money

    def __str__(self):
        return f"Player({self.ID}, {self.money})"

    def __repr__(self):
        return self.__str__()


class Tile:
    """
    Tile je generická trieda pre políčko.
    Políčko má pozíciu a vlastníka práva na výhybku.
    """

    position: Position
    right_of_switch: Union[int, None]

    def buildable(self, *args, **kwargs) -> bool:
        """
        🤮 (vracia) True ak sa dá na políčku stavať.
        """
        return False

    def max_trains(self) -> int:
        """
        🤮 (vracia) maximálny počet vlakov, ktoré môžu byť na políčku.
        """
        return 0

    def map_neighbour(
            self, map: dict[Position, "Tile"], direction
    ) -> Union["Tile", None]:
        """
        🤮 (vracia) susedné políčko vo vybranom smere.
        """
        pos = _direction_positions.get(direction)
        if pos:
            return map.get(self.position + pos, None)

    def all_neighbours(self, map: dict[Position, "Tile"]) -> set["Tile"]:
        """
        🤮 (vracia) množinu všetkých susedných políčok.
        """
        neighbours = set()
        for dir, pos in _direction_positions.items():
            if tile := map.get(self.position + pos, None):
                neighbours.add(tile)
        return neighbours

    def neighbours_from_direction(
            self, direction: Directions, map: dict[Position, "Tile"]
    ) -> set["Tile"]:
        """
        🤮 (vracia) množinu všetkých susedných políčok, na ktoré sa váš vlak dokáže dostať,
         keď prichádza z vybraného smeru.
        """
        return set()

    def dir_to_neighbour(self, other: "Tile") -> Directions:
        """
        🤮 (vracia) smer v ktorom je vybrané susedné políčko.
        """
        diff = other.position - self.position
        return _position_directions.get(diff, Directions.NONE)

    def reachable(self, map: dict[Position, "Tile"]) -> set["Tile"]:
        """
        🤮 (vracia) množinu susedných políčok, na ktoré sa dá z políčka dostať po trati, bez ohľadu na smer.
        """
        return self.all_neighbours(map)


class TileEmpty(Tile):
    def __init__(self, position: Position):
        self.position = position
        self.right_of_switch = None

    def buildable(self, *args, **kwargs) -> bool:
        return True

    def max_trains(self) -> int:
        return 0

    def __str__(self):
        return f"TileEmpty({self.position})"

    def __repr__(self):
        return self.__str__()


class TileWater(Tile):
    def __init__(self, position: Position):
        self.position = position
        self.right_of_switch = None

    def buildable(self, *args, **kwargs) -> bool:
        return False

    def max_trains(self) -> int:
        return 0

    def __str__(self):
        return f"TileWater({self.position})"

    def __repr__(self):
        return self.__str__()

    def reachable(self, map: dict[Position, Tile]):
        return set()


class TileTrack(Tile):
    def __init__(
            self, position: Position, player: int, segments: Tuple[Directions, Directions]
    ):
        self.position = position
        self.right_of_switch = player
        self.segments = (Directions(segments[0]), Directions(segments[1]))

    def buildable(self, other: Player) -> bool:
        return self.right_of_switch == other.ID

    def max_trains(self) -> int:
        return 1

    def __str__(self):
        return f"TileTrack({self.position}, {self.right_of_switch}, {self.segments})"

    def __repr__(self):
        return self.__str__()

    def neighbours_from_direction(
            self, direction: Directions, map: dict[Position, "Tile"]
    ) -> set["Tile"]:
        if direction.opposite() == self.segments[0]:
            return _tiles_to_set([self.map_neighbour(map, self.segments[1])])
        elif direction.opposite() == self.segments[1]:
            return _tiles_to_set([self.map_neighbour(map, self.segments[0])])
        return set()

    def reachable(self, map: dict[Position, "Tile"]):
        return _tiles_to_set(
            [
                self.map_neighbour(map, self.segments[0]),
                self.map_neighbour(map, self.segments[1]),
            ]
        )


class TileSwitch(Tile):
    def __init__(
            self,
            position: Position,
            player: int,
            is_switched: bool,
            straight_direction: Directions,
            switched_direction: Directions,
    ):
        self.position = position
        self.right_of_switch = player
        self.is_switched = is_switched
        self.straight_direction = Directions(straight_direction)
        self.switched_direction = Directions(switched_direction)

    def buildable(self, other: Player) -> bool:
        return self.right_of_switch == other.ID

    def max_trains(self) -> int:
        return 1

    def neighbours_from_direction(
            self, direction: Directions, map: dict[Position, "Tile"]
    ) -> set["Tile"]:
        if direction == self.straight_direction:
            neighbours = set()
            neighbours.add(self.map_neighbour(map, self.switched_direction))
            neighbours.add(self.map_neighbour(map, self.straight_direction.opposite()))
            return neighbours
        elif direction == self.straight_direction.opposite():
            return _tiles_to_set([self.map_neighbour(map, direction)])
        elif direction == self.switched_direction.opposite() and self.is_switched:
            return _tiles_to_set(
                [self.map_neighbour(map, self.straight_direction.opposite())]
            )
        return set()

    def __str__(self):
        return f"TileSwitch({self.position}, {self.is_switched}, {self.straight_direction}, {self.switched_direction})"

    def __repr__(self):
        return self.__str__()

    def reachable(self, map: dict[Position, Tile]) -> set[Tile]:
        return _tiles_to_set(
            [
                self.map_neighbour(map, self.straight_direction),
                self.map_neighbour(map, self.straight_direction.opposite),
                self.map_neighbour(map, self.switched_direction),
            ]
        )


class TileCrossing(Tile):
    def __init__(
            self, position: Position, player: int, directions: Tuple[Directions, Directions]
    ):
        self.position = position
        self.right_of_switch = player
        self.directions = (Directions(directions[0]), Directions(directions[1]))

    def buildable(self, other: Player) -> bool:
        return self.right_of_switch == other.ID

    def max_trains(self) -> int:
        return 1

    def neighbours_from_direction(
            self, direction: Directions, map: dict[Position, "Tile"]
    ) -> set["Tile"]:
        if direction == self.directions[0]:
            return _tiles_to_set([self.map_neighbour(map, direction)])
        elif direction == self.directions[1]:
            return _tiles_to_set([self.map_neighbour(map, direction)])
        elif direction == self.directions[0].opposite():
            return _tiles_to_set([self.map_neighbour(map, direction)])
        elif direction == self.directions[1].opposite():
            return _tiles_to_set([self.map_neighbour(map, direction)])
        return set()

    def __str__(self):
        return f"TileCrossing({self.position}, {self.directions})"

    def __repr__(self):
        return self.__str__()

    def reachable(self, map: dict[Position, Tile]) -> set[Tile]:
        return _tiles_to_set(
            [
                self.map_neighbour(map, self.directions[0]),
                self.map_neighbour(map, self.directions[0].opposite()),
                self.map_neighbour(map, self.directions[1]),
                self.map_neighbour(map, self.directions[1].opposite()),
            ]
        )


class TileDepot(Tile):
    def __init__(self, position: Position, player: int):
        self.position = position
        self.right_of_switch = player

    def buildable(self, other: Player) -> bool:
        return False

    def max_trains(self) -> int:
        return 10 ** 9 + 7

    def neighbours_from_direction(
            self, direction: Directions, map: dict[Position, "Tile"]
    ) -> set["Tile"]:
        return _tiles_to_set([self.map_neighbour(map, direction)])

    def __str__(self):
        return f"TileDepot({self.position}, {self.right_of_switch})"

    def __repr__(self):
        return self.__str__()


class Recipe:
    """
    Recept je trieda, ktorá reprezentuje výrobný proces suroviny. Má vstupné a výstupné materiály.
    """

    input: list[Materials]
    output: list[Materials]

    def __init__(self, input: list[Materials], output: list[Materials]):
        self.input = input
        self.output = output


class FactoryType(enum.Enum):
    """
    FactoryType je trieda, ktorá reprezentuje typ fabrík.
    """

    Forest = 0
    Farm = 1
    CoalMine = 2
    IronMine = 3
    SteelMill = 4
    SawMill = 5
    GoodsFactory = 6
    FoodFactory = 7
    Town = 8


RECIPES = {
    FactoryType.IronMine: Recipe([], [Materials.IronOre]),
    FactoryType.CoalMine: Recipe([], [Materials.Coal]),
    FactoryType.Forest: Recipe([], [Materials.Logs]),
    FactoryType.Farm: Recipe([], [Materials.Grain]),
    FactoryType.SteelMill: Recipe(
        [Materials.IronOre, Materials.Coal], [Materials.Steel]
    ),
    FactoryType.SawMill: Recipe([Materials.Logs], [Materials.Planks]),
    FactoryType.FoodFactory: Recipe([Materials.Grain], [Materials.Food]),
    FactoryType.GoodsFactory: Recipe(
        [Materials.Steel, Materials.Planks], [Materials.Goods]
    ),
    FactoryType.Town: Recipe([Materials.Goods, Materials.Food], []),
}


class TileFactory(Tile):
    def __init__(
            self,
            position: Position,
            straight_direction: Directions,
            type: FactoryType,
            cargo: dict[Materials, int],
    ):
        self.position = position
        self.straight_direction = Directions(straight_direction)
        self.type = FactoryType(type)
        self.cargo = dict[Materials, int](cargo)

    def buildable(self, other: Player) -> bool:
        return False

    def max_trains(self) -> int:
        return 10 ** 9 + 7

    def neighbours_from_direction(
            self, direction: Directions, map: dict[Position, "Tile"]
    ):
        if direction == self.straight_direction:
            return _tiles_to_set([self.map_neighbour(map, self.straight_direction)])
        elif direction == self.straight_direction.opposite():
            return _tiles_to_set(
                [self.map_neighbour(map, self.straight_direction.opposite())]
            )
        return set()

    def __str__(self):
        return f"TileFactory({self.position}, {self.straight_direction}, {self.type})"

    def __repr__(self):
        return self.__str__()

    def reachable(self, map: dict[Position, Tile]) -> set[Tile]:
        return _tiles_to_set(
            [
                self.map_neighbour(map, self.straight_direction),
                self.map_neighbour(map, self.straight_direction.opposite()),
            ]
        )


class Train:
    """
    Train je trieda, ktorá reprezentuje Vlak.
    O vlakoch vieme povedať: ich id, vlastníka, pozíciu, smer, počet vagónov, prevážanú surovinu a jej množstvo.
    """

    id: int
    player: int
    position: Position
    direction: Directions
    wagons: int
    cargo_amount: int
    cargo_material: Materials

    def __init__(self, data):
        self.id = data.get("id")
        self.player = data.get("player")
        self.position = Position.from_json(data.get("position"))
        self.direction = Directions(data.get("direction"))
        self.wagons = data.get("wagons")
        self.cargo_amount = data.get("cargo_amount")
        self.cargo_material = Materials(data.get("cargo_material"))

    def __str__(self):
        return f"Train({self.id}, {self.player}, {self.position}, {self.direction}, {self.wagons}, {self.cargo_amount})"

    def __repr__(self):
        return self.__str__()


class Turn:
    name: str
    data: property

    def to_dict(self):
        return {"turn": self.name, "data": self.data}


class TurnBuild(Turn):
    """
    TurnBuild reprezentuje ťah stavania trate.
    """

    name = "build"

    def __init__(self, position: Position, segments: list[Directions]):
        """
        Pre stavanie potrebujeme zvoliť pozíciu políčka na mape a zoznam segmentov, ktorý chceme stavať.
        """
        self.position = position
        self.segments = list(map(Directions, segments))

    @property
    def data(self) -> dict:
        return {
            "position": self.position.to_dict(),
            "segment": [segment.value for segment in self.segments],
        }


class TurnBuildDepot(Turn):
    """
    TurnBuild reprezentuje ťah stavania depa.
    """

    name = "build_depot"

    def __init__(self, position: Position):
        """
        Pre stavanie potrebujeme zvoliť pozíciu políčka na mape.
        """
        self.position = position

    @property
    def data(self) -> dict:
        return {"position": self.position.to_dict()}

    def __str__(self):
        return f"TurnBuildDepot({self.position})"

    def __repr__(self):
        return self.__str__()


class TurnToggleSwitch(Turn):
    """
    TurnToggleSwitch reprezentuje ťah prehodenie výhybku.
    """

    name = "toggle_switch"

    def __init__(self, position: Position):
        """
        Na prehodenie výhybky potrebe zvoliť pozíciu políčka s výhybkou.
        """
        self.position = position

    @property
    def data(self) -> dict:
        return {"position": self.position.to_dict()}

    def __str__(self):
        return f"TurnToggleSwitch({self.position})"

    def __repr__(self):
        return self.__str__()


class TurnDestroy(Turn):
    """
    TurnDestroy reprezentuje ťah zničenia trate. Zničí všetko a na políčku bude TileEmpty. Keď stojí vlak na políčku, tak sa políčko zničiť nedá.
    """

    name = "destroy"

    def __init__(self, position: Position):
        """
        Na zničenie potrebujeme zvoliť pozíciu políčka.
        """
        self.position = position

    @property
    def data(self) -> dict:
        return {"position": self.position.to_dict()}

    def __str__(self):
        return f"TurnDestroy({self.position})"

    def __repr__(self):
        return self.__str__()


class TurnTrainDirection(Turn):
    """
    TurnTrainDirection reprezentuje ťah zmeny smeru vlaku. Taktiež ide začatie pohybu. Za jazdy smer však meniť nevieme.
    """

    name = "train_direction"

    def __init__(self, train: Train, direction: Directions):
        """
        Pre zmenu smeru vlaku potrebujeme zvoliť vlak a smer, do ktorého chceme aby šiel.
        """
        self.train = train
        self.direction = direction

    @property
    def data(self) -> dict:
        return {"train": self.train.id, "direction": self.direction.value}

    def __str__(self):
        return f"TurnTrainDirection({self.train}, {self.direction})"

    def __repr__(self):
        return self.__str__()


class TurnTrainStop(Turn):
    """
    TurnToggleSwitch reprezentuje ťah zastavenia vlaku.
    """

    name = "train_stop"

    def __init__(self, train: Train):
        """
        Na zastavenie vlaku potrebujeme zvoliť vlak, ktorý chceme zastaviť.
        """
        self.train = train

    @property
    def data(self) -> dict:
        return {"train": self.train.id}

    def __str__(self):
        return f"TurnTrainStop({self.train})"

    def __repr__(self):
        return self.__str__()


class TurnTrainLoad(Turn):
    """
    TurnTrainLoad reprezentuje ťah nakládky vlaku. Náš vlak pri tom musí stáť vo fabrike.
    """

    name = "train_load"

    def __init__(self, train: Train, cargo_amount: int, cargo_material: Materials):
        """
        Pre vykládku potrebujeme zvoliť vlak, ktorý sa má naložiť, počet suroviny, ktorú chceme vo fabrike vyložiť
         a taktiež o akú surovinu ide.
        """
        self.train = train
        self.cargo_amount = cargo_amount
        self.cargo_material = cargo_material

    @property
    def data(self) -> dict:
        return {
            "train": self.train.id,
            "cargo_amount": self.cargo_amount,
            "cargo_material": self.cargo_material,
        }

    def __str__(self):
        return f"TurnTrainLoad({self.train}, {self.cargo_amount}, {self.cargo_material})"

    def __repr__(self):
        return self.__str__()


class TurnTrainUnload(Turn):
    """
    TurnTrainUnload reprezentuje ťah vykládky vlaku. Náš vlak pri tom musí stáť vo fabrike.
    """

    name = "train_unload"

    def __init__(self, train: Train, cargo_amount: int):
        """
        Pre vykládku potrebujeme zvoliť vlak, ktorý sa má vyložiť a počet suroviny, ktorú chceme vo fabrike vyložiť.
        """
        self.train = train
        self.cargo_amount = cargo_amount

    @property
    def data(self) -> dict:
        return {"train": self.train.id, "cargo_amount": self.cargo_amount}

    def __str__(self):
        return f"TurnTrainUnload({self.train}, {self.cargo_amount})"

    def __repr__(self):
        return self.__str__()


class TurnTrainBuy(Turn):
    """
    TurnTrainBuy reprezentuje ťah nákupu vlaku.
    """

    name = "train_buy"

    def __init__(self, depot: Position):
        """
        Pre nákup vlaku potrebujeme zvoliť niektoré naše depo, v ktorom sa vytvorí.
        """
        self.depot = depot

    @property
    def data(self) -> dict:
        return {"depot": self.depot.to_dict()}

    def __str__(self):
        return f"TurnTrainBuy({self.depot})"

    def __repr__(self):
        return self.__str__()


class TurnTrainAddWagon(Turn):
    """
    TurnTrainAddWagon reprezentuje ťah nákupu vagónu. Vlak pri tom musí stáť niektorom našom depe.
    """

    name = "train_add_wagon"

    def __init__(self, train: Train):
        """
        Pre nákup vagónu musíme zvoliť vlak, do ktorého chceme pridať vagón.
        """
        self.train = train

    @property
    def data(self) -> dict:
        return {"train": self.train.id}

    def __str__(self):
        return f"TurnTrainAddWagon({self.train})"

    def __repr__(self):
        return self.__str__()


class TurnHorn(Turn):
    """
    TurnHorn reprezentuje ťah trúbenia vlaku. Máme na výber z dvoch vlastných mp3.
    """

    name = "horn"

    def __init__(self, id: int):
        """
        Na zatrúbenie potrebujeme zvoliť id mp3ky.
        """
        self.id = id

    @property
    def data(self) -> dict:
        return {"id": self.id}

    def __str__(self):
        return f"TurnHorn({self.id})"

    def __repr__(self):
        return self.__str__()


class Game:
    """
    Game je trieda, ktorá preprezenuje celú hru.

    * map - objekt Mapy
    * player - váš hráč
    * enemy_player - zoznam neprateľských hráčov, ktorých vidíte
    * items - zoznam všetkých vecí, ktoré vidíte
    """

    player: Player
    color: str
    map: dict[Position, Tile]
    radius: int

    def __init__(self, color):
        self.color = color
        self.map = dict()
        self.factories = dict()
        self.trains = []
        self.my_trains = []
        self.my_depos = dict()

    @staticmethod
    def log(*args):
        """
        Vypíše dáta do logu. Syntax je rovnaká ako print().
        """
        print(*args, file=sys.stderr)

    def fill_with_empty(self):
        if len(self.map) > self.radius:
            return
        for r in range(-self.radius, self.radius + 1):
            minQ = -self.radius - r if r < 0 else -self.radius
            maxQ = self.radius - r if r > 0 else self.radius
            for q in range(minQ, maxQ + 1):
                s = -q - r
                pos = Position(q, r, s)
                self.map[pos] = TileEmpty(pos)

    def _read_state(self):
        data = json.loads(input())
        self.radius = data.get("map_radius")
        self.trains = []
        self.my_trains = []
        self.player = Player(
            data.get("player").get("id"), data.get("player").get("money")
        )
        self.fill_with_empty()

        for tile_data in data.get("dirty_tiles"):
            type = tile_data.get("type")
            tile = tile_data.get("data")
            pos = Position.from_json(tile.get("position"))
            player_id = tile.get("right_of_switch")
            if type == "track":
                self.map[pos] = TileTrack(pos, player_id, tile.get("segments"))
            elif type == "switch":
                self.map[pos] = TileSwitch(
                    pos,
                    player_id,
                    tile.get("is_switched"),
                    tile.get("straight_direction"),
                    tile.get("switched_direction"),
                )
            elif type == "crossing":
                self.map[pos] = TileCrossing(pos, player_id, tile.get("directions"))
            elif type == "depot":
                self.map[pos] = TileDepot(pos, player_id)
                if self.player.ID == player_id:
                    self.my_depos[pos] = TileDepot(pos, player_id)
            elif type == "factory":
                self.map[pos] = TileFactory(
                    pos,
                    tile.get("straight_direction"),
                    tile.get("factory_type"),
                    tile.get("cargo"),
                )
                self.factories[pos] = TileFactory(
                    pos,
                    tile.get("straight_direction"),
                    tile.get("factory_type"),
                    tile.get("cargo"),
                )
            elif type == "empty":
                self.map[pos] = TileEmpty(pos)
            elif type == "water":
                self.map[pos] = TileWater(pos)
        for train in data.get("trains"):
            t = Train(train)
            if t.player == self.player.ID:
                self.my_trains.append(t)
            self.trains.append(t)

        input()

    def greet(self):
        print(json.dumps({"color": self.color}))
        print(".")

    def make_turns(self) -> list[Turn]:
        raise NotImplementedError()

    def _send_turns(self, turns: list[Turn]):
        _turns = []
        for turn in turns:
            _turns.append(turn.to_dict())
        _turns = json.dumps(_turns)
        print(_turns)
        print(".")

    def run(self):
        self.greet()
        while True:
            self._read_state()
            turns = self.make_turns()
            self._send_turns(turns)
